package checkers.nonnull;

import java.io.File;
import java.io.IOException;
import java.lang.annotation.Annotation;
import java.util.Properties;

import javax.annotation.processing.ProcessingEnvironment;

import checkers.quals.NonNull;
import checkers.subtype.SubtypeChecker;
import checkers.types.AnnotatedTypeFactory;

import com.sun.source.tree.CompilationUnitTree;

public class NonnullChecker extends SubtypeChecker {

    @Override
    protected Class<? extends Annotation> getCheckedAnnotation() {
        return NonNull.class;
    }

    @Override
    protected Properties getMessages(String fileName) throws IOException {
        if (new File(fileName).exists())
            return super.getMessages("nonnull.properties");
        
        Properties msgDefaults = new Properties();
        msgDefaults.setProperty("assignment.invalid", "invalid assignment type");
        msgDefaults.setProperty("argument.invalid", "invalid argument type");
        msgDefaults.setProperty("receiver.invalid", "invalid receiver type");
        msgDefaults.setProperty("return.invalid", "invalid return type");
        msgDefaults.setProperty("override.param.invalid", "invalid overriding parameter type");
        msgDefaults.setProperty("override.return.invalid", "invalid overriding return type");
        
        return msgDefaults;
    }

    @Override
    public AnnotatedTypeFactory getFactory(ProcessingEnvironment env, CompilationUnitTree root) {
        return new NonnullAnnotatedTypeFactory(env, root);
    }
}
