package checkers.types;

import javax.lang.model.element.Element;

import com.sun.source.tree.*;
import com.sun.source.util.*;
import com.sun.tools.javac.code.*;
import com.sun.tools.javac.tree.*;
import com.sun.tools.javac.tree.JCTree.JCMethodInvocation;

// TODO *** clean up and document

/**
 * Static utility methods used by annotation abstractions in this package.
 */
public class InternalUtils {
    
    /** Cannot be instantiated. */
    private InternalUtils() {
    }
       
    /**
     * Gets the enclosing method of the tree node defined by the given
     * {@code {@link TreePath}}.
     * 
     * @param path
     *            the path defining the tree node
     * @return the enclosing method as given by the path, or null if one does
     *         not exist
     */
    public static Tree enclosingMethod(final TreePath path) {

        TreePath p = path; 
        while (p != null && p.getParentPath() != null) {
            Tree leaf = p.getLeaf();
            if (leaf.getKind() == Tree.Kind.METHOD)
                return leaf;
            p = p.getParentPath();
        }

        return null;
    }
        
    public static Element enclosingSymbol(final TreePath path) {
               
        Tree enclosingMethod = enclosingMethod(path);
        if (enclosingMethod != null) {
            Element elt = symbol(enclosingMethod);
            return elt;
        }
        
        TreePath p = path;
        Tree leaf = null;
        while (p != null && p.getParentPath() != null) {
            leaf = p.getLeaf();
            if (leaf.getKind() == Tree.Kind.VARIABLE || leaf.getKind() == Tree.Kind.CLASS)
                break;
            p = p.getParentPath();
        }
        
        if (leaf == null) 
            return null;
        else if (leaf.getKind() == Tree.Kind.CLASS)
            return TreeInfo.symbolFor((JCTree)leaf);
        
        return symbol(leaf);
    }
    
    public static Element enclosingAnnotationSource(Trees trees, final TreePath path) {
        Tree tree;
        TreePath fpath = path;
        while (fpath != null && fpath.getLeaf() != null) {
            tree = fpath.getLeaf();
            if (tree.getKind() == Tree.Kind.BLOCK) {
                BlockTree bt = (BlockTree)tree;
                if (bt.isStatic())
                    return symbol(fpath.getParentPath().getLeaf());
            }
            fpath = fpath.getParentPath();
        }
        return enclosingSymbol(path);
    }
    
    public static boolean refersTo(Tree source, Tree dest) {
        
        if (!(source instanceof JCTree) || !(dest instanceof JCTree))
            return false;
        
        return (source == dest || ((JCTree)source).pos == ((JCTree)dest).pos);
    }
    
    /**
     * Gets the javac-internal {@code {@link Symbol}} for the given Tree API
     * node.
     *
     * @param tree
     *            the tree node to get the symbol for
     * @throws IllegalArgumentException
     *             if {@code tree} is null or is not a valid javac-internal tree
     * @return the {@code {@link Symbol}} for the given tree, or null if one
     *         could not be found
     */
    public static Symbol symbol(Tree tree) {
        
        if (tree == null)
            throw new IllegalArgumentException("tree is null");
        
        if (!(tree instanceof JCTree))
            throw new IllegalArgumentException("tree is not a valid Javac tree");
        
        if (tree instanceof VariableTree || tree instanceof MethodTree)
            return TreeInfo.symbolFor((JCTree) tree);
        
        // symbol() only works on MethodSelects, so we need to get it manually
        // for method invocations.
        if (tree instanceof MethodInvocationTree)
            return TreeInfo.symbol(((JCMethodInvocation) tree)
            .getMethodSelect());
        
        return TreeInfo.symbol((JCTree) tree);
    }
    
    public static boolean isAnonymousConstructor(TreePath path) {
        
        Element e = InternalUtils.enclosingSymbol(path);
        if (e == null)
            return false;
        
        if ((((Symbol)e).flags() & Flags.ANONCONSTR) != 0)
            return true;
        
        return false;
    }
}
